# Copyright 2020 Nicene Nerd <macadamiadaze@gmail.com>
# Licensed under GPLv3+
from pathlib import Path
from typing import List, Union, Dict
from zlib import crc32

import oead

from bcml import util, mergers
from bcml.util import BcmlMod
from bcml import bcml as rsext


def get_stock_actorinfo() -> oead.byml.Hash:
    actorinfo = util.get_game_file("Actor/ActorInfo.product.sbyml")
    return oead.byml.from_binary(util.decompress(actorinfo.read_bytes()))


class ActorInfoMerger(mergers.Merger):
    NAME: str = "actors"

    def __init__(self):
        super().__init__(
            "actor info",
            "Merges changes to ActorInfo.product.byml",
            "actorinfo.yml",
            {},
        )

    @util.timed
    def generate_diff(self, mod_dir: Path, modded_files: List[Union[Path, str]]):
        try:
            actor_file: Path = next(
                iter(
                    [
                        file
                        for file in modded_files
                        if isinstance(file, Path)
                        and file.name == "ActorInfo.product.sbyml"
                    ]
                )
            )
        except StopIteration:
            return None
        return rsext.mergers.actorinfo.diff_actorinfo(str(actor_file))

    def log_diff(self, mod_dir: Path, diff_material: Union[oead.byml.Hash, list]):
        if isinstance(diff_material, List):
            diff_material = self.generate_diff(mod_dir, diff_material)
        if diff_material:
            (mod_dir / "logs" / self._log_name).write_bytes(diff_material)

    def get_mod_diff(self, mod: BcmlMod):
        diffs: Dict[str, oead.Byml.Hash] = {}
        if self.is_mod_logged(mod):
            util.dict_merge(
                diffs,
                oead.byml.from_text(
                    (mod.path / "logs" / self._log_name).read_text(encoding="utf-8")
                ),
                overwrite_lists=True,
            )
        for opt in {d for d in (mod.path / "options").glob("*") if d.is_dir()}:
            if (opt / "logs" / self._log_name).exists():
                util.dict_merge(
                    diffs,
                    oead.byml.from_text(
                        (opt / "logs" / self._log_name).read_text("utf-8")
                    ),
                    overwrite_lists=True,
                )
        return diffs

    def get_all_diffs(self):
        diffs = []
        for mod in util.get_installed_mods():
            diffs.append(self.get_mod_diff(mod))
        return diffs

    def consolidate_diffs(self, diffs: list):
        all_diffs: Dict[str, oead.Byml.Hash] = {}
        inst_sizes = {}
        for diff in diffs:
            for actor_hash, inst_size in [(h, a["instSize"].v) for h, a in diff.items() if "instSize" in a]:
                if inst_size > inst_sizes.get(actor_hash, 0):
                    inst_sizes[actor_hash] = inst_size
            util.dict_merge(all_diffs, diff, overwrite_lists=True)
        for actor_hash, inst_size in inst_sizes.items():
            all_diffs[actor_hash]["instSize"] = oead.S32(inst_size)
        return oead.byml.Hash(all_diffs)

    @util.timed
    def perform_merge(self):
        actor_path = (
            util.get_master_modpack_dir()
            / util.get_content_path()
            / "Actor"
            / "ActorInfo.product.sbyml"
        )
        print("Loading modded actor info...")
        modded_actors = self.consolidate_diffs(self.get_all_diffs())
        if not modded_actors:
            print("No actor info merging necessary.")
            if actor_path.exists():
                actor_path.unlink()
            return

        bin_data = oead.byml.to_binary(modded_actors, False)
        rsext.mergers.actorinfo.merge_actorinfo(bin_data)
        print("Actor info merged successfully")

    def get_checkbox_options(self):
        return []

    def get_mod_edit_info(self, mod: util.BcmlMod) -> set:
        log = {int(k): v for k, v in self.get_mod_diff(mod).items()}
        actors = get_stock_actorinfo()
        stock_names = {
            int(actors["Hashes"][i]): actors["Actors"][i]["name"]
            for i in range(len(actors["Hashes"]))
        }
        return {
            (stock_names[actor] if actor in stock_names else log[actor]["name"])
            for actor in log
        }



import rstb
import oead

from functools import lru_cache
from pathlib import Path
from typing import List, Union
from bcml import mergers, util
from bcml.util import BcmlMod
from bcml.mergers import rstable


def get_stock_areadata() -> oead.byml.Hash:
    if not hasattr(get_stock_areadata, "areadata"):
        get_stock_areadata.areadata = oead.byml.to_text(
            oead.byml.Hash(
                {
                    str(area["AreaNumber"].v): area
                    for area in oead.byml.from_binary(
                        util.get_nested_file_bytes(
                            str(util.get_game_file("Pack/Bootup.pack"))
                            + "//Ecosystem/AreaData.sbyml",
                            unyaz=True,
                        )
                    )
                }
            )
        )
    return oead.byml.from_text(get_stock_areadata.areadata)


def get_modded_areadata(areadata: oead.byml.Array) -> oead.byml.Hash:
    stock_areadata = get_stock_areadata()
    mod_areadata = oead.byml.Hash(
        {str(area["AreaNumber"].v): area for area in areadata}
    )
    modified = oead.byml.Hash()
    try:
        for area_num in [
            k for k in mod_areadata if mod_areadata[k] != stock_areadata[k]
        ]:
            modified[area_num] = oead.byml.Hash(
                {
                    k: v
                    for k, v in mod_areadata[area_num].items()
                    if v != stock_areadata[area_num][k]
                }
            )
    except KeyError as err:
        raise RuntimeError(
            f"Invalid AreaData.sbyml. One or more areas missing key: {err.args}"
        )
    return modified


class AreaDataMerger(mergers.Merger):
    NAME: str = "areadata"

    def __init__(self):
        super().__init__(
            "area data",
            "Merges changes to AreaData.byml",
            "areadata.yml",
            options={},
        )

    def generate_diff(self, mod_dir: Path, modded_files: List[Union[Path, str]]):
        if (
            f"{util.get_content_path()}/Pack/Bootup.pack//Ecosystem/AreaData.sbyml"
            in modded_files
        ):
            print("Logging modded areadata...")
            bootup_sarc = oead.Sarc(
                (
                    mod_dir / util.get_content_path() / "Pack" / "Bootup.pack"
                ).read_bytes()
            )
            areadata = oead.byml.from_binary(
                util.decompress(bootup_sarc.get_file("Ecosystem/AreaData.sbyml").data)
            )
            diff = get_modded_areadata(areadata)
            del bootup_sarc
            del areadata
            return diff
        else:
            return {}

    def log_diff(self, mod_dir: Path, diff_material):
        if isinstance(diff_material, List):
            diff_material = self.generate_diff(mod_dir, diff_material)
        if diff_material:
            (mod_dir / "logs" / self._log_name).write_text(
                oead.byml.to_text(diff_material), encoding="utf-8"
            )
            del diff_material

    def get_mod_diff(self, mod: BcmlMod):
        diffs = {}
        if self.is_mod_logged(mod):
            util.dict_merge(
                diffs,
                oead.byml.from_text(
                    (mod.path / "logs" / self._log_name).read_text(encoding="utf-8")
                ),
                overwrite_lists=True,
            )
        for opt in {d for d in (mod.path / "options").glob("*") if d.is_dir()}:
            if (opt / "logs" / self._log_name).exists():
                util.dict_merge(
                    diffs,
                    oead.byml.from_text(
                        (opt / "logs" / self._log_name).read_text("utf-8")
                    ),
                    overwrite_lists=True,
                )
        return diffs

    def get_all_diffs(self):
        diffs = []
        for mod in util.get_installed_mods():
            diff = self.get_mod_diff(mod)
            if diff:
                diffs.append(diff)
        return diffs

    def consolidate_diffs(self, diffs: list):
        if not diffs:
            return {}
        all_diffs = oead.byml.Hash()
        for diff in diffs:
            util.dict_merge(all_diffs, diff, overwrite_lists=True)
        return all_diffs

    @util.timed
    def perform_merge(self):
        merged_areadata = util.get_master_modpack_dir() / "logs" / "areadata.byml"
        areadata_merge_log = util.get_master_modpack_dir() / "logs" / "areadata.log"

        print("Loading area data mods...")
        modded_areadata = self.consolidate_diffs(self.get_all_diffs())
        areadata_mod_hash = hash(str(modded_areadata))
        if not modded_areadata:
            print("No area data merging necessary")
            if merged_areadata.exists():
                merged_areadata.unlink()
                areadata_merge_log.unlink()
                try:
                    stock_areadata = util.get_nested_file_bytes(
                        (
                            str(util.get_game_file("Pack/Bootup.pack"))
                            + "//Ecosystem/AreaData.sbyml"
                        ),
                        unyaz=False,
                    )
                    util.inject_file_into_sarc(
                        "Ecosystem/AreaData.sbyml",
                        stock_areadata,
                        "Pack/Bootup.pack",
                    )
                except FileNotFoundError:
                    pass
            return
        if (
            areadata_merge_log.exists()
            and areadata_merge_log.read_text() == areadata_mod_hash
        ):
            print("No area data merging necessary")
            return

        new_areadata = get_stock_areadata()
        util.dict_merge(new_areadata, modded_areadata, overwrite_lists=True)

        print("Writing new area data...")
        areadata_bytes = oead.byml.to_binary(
            oead.byml.Array(
                [v for _, v in sorted(new_areadata.items(), key=lambda x: int(x[0]))]
            ),
            big_endian=util.get_settings("wiiu"),
        )
        del new_areadata
        util.inject_file_into_sarc(
            "Ecosystem/AreaData.sbyml",
            util.compress(areadata_bytes),
            "Pack/Bootup.pack",
            create_sarc=True,
        )
        print("Saving area data merge log...")
        areadata_merge_log.parent.mkdir(parents=True, exist_ok=True)
        areadata_merge_log.write_text(str(areadata_mod_hash))
        merged_areadata.write_bytes(areadata_bytes)

        print("Updating RSTB...")
        rstb_size = rstb.SizeCalculator().calculate_file_size_with_ext(
            bytes(areadata_bytes), True, ".byml"
        )
        del areadata_bytes
        rstable.set_size("Ecosystem/AreaData.byml", rstb_size)

    def get_checkbox_options(self):
        return []

    @staticmethod
    def is_bootup_injector():
        return True

    def get_bootup_injection(self):
        tmp_sarc = util.get_master_modpack_dir() / "logs" / "areadata.byml"
        if tmp_sarc.exists():
            return (
                "Ecosystem/AreaData.sbyml",
                util.compress(tmp_sarc.read_bytes()),
            )
        else:
            return

    def get_mod_edit_info(self, mod: util.BcmlMod) -> set:
        return self.get_mod_diff(mod).keys()



"""Provides functions for diffing and merging BotW game text files"""
# Copyright 2020 Nicene Nerd <macadamiadaze@gmail.com>
# Licensed under GPLv3+
import json
import multiprocessing
import subprocess
from functools import partial, lru_cache
from pathlib import Path
from platform import system
from tempfile import TemporaryDirectory, NamedTemporaryFile
from typing import List, Union, Set, ByteString

import oead
import xxhash

from bcml import mergers, util
from bcml import bcml as rsext
from bcml.util import get_7z_path

LANGUAGES = [
    "USen",
    "EUen",
    "USfr",
    "USes",
    "EUde",
    "EUes",
    "EUfr",
    "EUit",
    "EUnl",
    "EUru",
    "CNzh",
    "JPja",
    "KRko",
    "TWzh",
]


def swap_region(mod_pack: Path, user_lang: str) -> Path:
    mod_sarc = oead.Sarc(mod_pack.read_bytes())
    mod_msg_data = mod_sarc.get_file(0).data
    new_pack = oead.SarcWriter(
        endian=oead.Endianness.Big
        if util.get_settings("wiiu")
        else oead.Endianness.Little
    )
    new_pack.files[f"Message/Msg_{user_lang}.product.ssarc"] = oead.Bytes(mod_msg_data)
    new_pack_path = mod_pack.with_name(f"Bootup_{user_lang}.temppack")
    new_pack_path.write_bytes(new_pack.write()[1])
    return new_pack_path


def map_languages(dest_langs: Set[str], src_langs: Set[str]) -> dict:
    """
    Function to map languages from one set to another

    Parameter: dest_langs: Set[str] - The languages to be mapped to
    - For diffing: the languages our mod has packs for, to diff with
    - For merging: the languages the user has, that we want to merge into

    Parameter: src_langs: Set[str] - The languages that will serve as the source
    - For diffing: the languages our dumps have packs for, to diff against
    - For merging: the languages our mod has logs for, to merge with

    Returns: Dict[str, str] - A dict where the keys are from dest_lang,
        and the values are the languages most closely mapped to them
    """
    lang_map: dict = {}
    for dest_lang in dest_langs:
        if dest_lang in src_langs:
            lang_map[dest_lang] = dest_lang
        else:
            for src_lang in [l for l in src_langs if l[2:4] == dest_lang[2:4]]:
                lang_map[dest_lang] = src_lang
                break
            if dest_lang in lang_map:
                continue
            for src_lang in [l for l in src_langs if l[2:4] == "en"]:
                lang_map[dest_lang] = src_lang
                break
            if dest_lang in lang_map:
                continue
            lang_map[dest_lang] = next(iter(src_langs))
    return lang_map


class TextsMerger(mergers.Merger):
    # pylint: disable=abstract-method
    """A merger for game texts"""
    NAME: str = "texts"

    def __init__(self, all_langs: bool = False):
        super().__init__(
            "game texts",
            "Merges changes to game texts",
            "texts.json",
            options={"all_langs": all_langs},
        )

    def generate_diff(self, mod_dir: Path, modded_files: List[Union[str, Path]]):
        print("Checking for modified languages...")
        mod_langs = {
            util.get_file_language(file)
            for file in modded_files
            if (
                isinstance(file, Path)
                and "Bootup_" in file.name
                and "Graphic" not in file.name
            )
        }
        if not mod_langs:
            return None
        util.vprint(f'Languages: {",".join(mod_langs)}')

        # find a user lang for each mod lang
        language_map = map_languages(mod_langs, util.get_user_languages())
        util.vprint("Language map:")
        util.vprint(language_map)

        language_diffs = {}
        for mod_lang, user_lang in language_map.items():
            print(f"Logging text changes for {user_lang}...")
            mod_pack = (
                mod_dir / util.get_content_path() / "Pack" / f"Bootup_{mod_lang}.pack"
            )
            if not user_lang == mod_lang:
                mod_pack = swap_region(mod_pack, user_lang)
            ref_pack = util.get_game_file(f"Pack/Bootup_{user_lang}.pack")
            language_diffs[user_lang] = rsext.mergers.texts.diff_language(
                str(mod_pack), str(ref_pack), user_lang[2:4] != mod_lang[2:4]
            )
            if not user_lang == mod_lang:
                mod_pack.unlink()

        return language_diffs

    def log_diff(self, mod_dir: Path, diff_material):
        if isinstance(diff_material, List):
            diff_material = self.generate_diff(mod_dir, diff_material)
        if diff_material:
            (mod_dir / "logs" / self._log_name).write_text(
                json.dumps(diff_material, ensure_ascii=False, indent=2),
                encoding="utf-8",
            )

    def get_mod_diff(self, mod: util.BcmlMod):
        diff = {}
        if self.is_mod_logged(mod):
            util.dict_merge(
                diff,
                json.loads((mod.path / "logs" / self._log_name).read_text("utf-8")),
            )
        for opt in {d for d in (mod.path / "options").glob("*") if d.is_dir()}:
            if (opt / "logs" / self._log_name).exists():
                util.dict_merge(
                    diff,
                    json.loads((opt / "logs" / self._log_name).read_text("utf-8")),
                    overwrite_lists=True,
                )
        return diff

    def get_mod_edit_info(self, mod: util.BcmlMod) -> set:
        diffs = set()
        if self.is_mod_logged(mod):
            for files in self.get_mod_diff(mod).values():
                diffs |= set(files.keys())
        return diffs

    def get_all_diffs(self):
        diffs = []
        for mod in util.get_installed_mods():
            diff = self.get_mod_diff(mod)
            if diff:
                diffs.append(diff)
        return diffs

    def consolidate_diffs(self, diffs: list):
        if not diffs:
            return {}
        main_diff = diffs[0]
        for diff in diffs[1:]:
            for lang, content in diff.items():
                if lang not in main_diff:
                    main_diff[lang] = content
                else:
                    for file, entries in content.items():
                        if file not in main_diff[lang]:
                            main_diff[lang][file] = entries
                        else:
                            for entry, msg in entries.items():
                                main_diff[lang][file][entry] = msg
        return main_diff

    @util.timed
    def perform_merge(self):
        # pylint: disable=unsupported-assignment-operation
        user_langs = (
            {util.get_settings("lang")}
            if not self._options["all_langs"]
            else util.get_user_languages()
        )
        print("Loading text mods...")
        diffs = self.consolidate_diffs(self.get_all_diffs())
        if not diffs:
            print("No text merge necessary")
            for bootup in util.get_master_modpack_dir().rglob("**/Bootup_????.pack"):
                bootup.unlink()
            return

        # find a mod lang for each user lang
        lang_map = map_languages(user_langs, set(diffs.keys()))
        for user_lang, mod_lang in lang_map.items():
            print(f"Merging modded texts for {mod_lang} into {user_lang}...")
            rsext.mergers.texts.merge_language(
                json.dumps(diffs[mod_lang]),
                str(util.get_game_file(f"Pack/Bootup_{user_lang}.pack")),
                str(
                    util.get_master_modpack_dir()
                    / util.get_content_path()
                    / "Pack"
                    / f"Bootup_{user_lang}.pack"
                ),
                util.get_settings("wiiu"),
            )
            print(f"{user_lang} texts merged successfully")

    def get_checkbox_options(self) -> List[tuple]:
        return [
            ("all_langs", "Merge texts for all game languages"),
        ]



from pathlib import Path
import sys

version = sys.argv[1]
major, minor, patch = version.split(".")

setup = Path("setup.py")
setup_text = setup.read_text().splitlines()
for i, line in enumerate(setup_text):
    if "version=" in line:
        setup_text[i] = f'    version="{version}",'
setup.write_text("\n".join(setup_text))

version_file = Path("bcml/__version__.py")
version_file_text = version_file.read_text().splitlines()
version_file_text[0:3] = [f"_MAJOR={major}", f"_MINOR={minor}", f'_PATCH="{patch}"']
version_file.write_text("\n".join(version_file_text))

cargo = Path("Cargo.toml")
cargo_text = cargo.read_text().splitlines()
for i, line in enumerate(cargo_text):
    if line.startswith("version"):
        cargo_text[i] = f'version = "{version}"'
cargo.write_text("\n".join(cargo_text))

pyproject = Path("pyproject.toml")
pyproject_text = pyproject.read_text().splitlines()
for i, line in enumerate(pyproject_text):
    if line.startswith("version"):
        pyproject_text[i] = f'version = "{version}"'
pyproject.write_text("\n".join(pyproject_text))

